/*
 * Copyright (c) 1998 ObjecTime Limited.  All rights reserved.
 * Unpublished--rights reserved under all copyright laws including
 * Copyright Laws of the United States.
 * 
 * ObjecTime (and logo) is a registered trademark of ObjecTime Limited.
 * Developer is a trademark of ObjecTime Limited.
 * 
 * "ObjecTime" shall mean the ObjecTime, SimulationRTS (InternalRTS),
 * EmulationRTS and TargetRTS (MicroRTS) computer programs.
 * 
 * The placing of the copyright notices on ObjecTime source code is not
 * intended to mean that they are published works, but are included in the
 * event of inadvertent publication. The source code of ObjecTime contains
 * trade secrets and other valuable confidential information of ObjecTime
 * Limited. ObjecTime Limited and its licensors retain ownership to
 * ObjecTime in both source and object code format, and related documentation
 * and all rights therein including copyrights, patents and trade secrets.
 * 
 *                        Restricted Rights Legend
 * 
 * With respect to any copies of the ObjecTime software and related
 * documentation provided to any U.S. Government department or agency, use,
 * duplication, or disclosure by the Government is subject to restrictions
 * as set forth in subparagraphs (a) through (d) of the Commercial Computer
 * Software-Restricted Rights clause FAR 52.227-19 and its successors. For
 * units of the Department of Defense (DoD), the license for this software is
 * subject to the "Restricted Rights" as that term is defined in the DFAR
 * 252.227-7013 (c) (1) (ii), Rights in Technical Data and Computer Software,
 * and its successors. Unpublished -- rights reserved under the Copyright
 * Laws of the United States.
 * 
 *                        ObjecTime Limited
 *                        340 March Road
 *                        Kanata, Ontario K2K 2E4
 *                        Canada
 * 
 * Unless otherwise agreed to in writing, use of ObjecTime is governed by
 * the Development License Agreement provided with the product.
 */

/*
 * This file contains source code to convert integer and real numbers into
 * their internal memory representation and to provide the corresponding
 * injection code for an RTByteBlock.
 *
 * This code was tested under the following conditions:
 *      - Microsoft Windows NT + Microsoft Visual-C++ 5.0
 *      - Sun Solaris 5.5.1 + gcc 2.7
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/******************************************************************************
 * Local definitions
 ******************************************************************************/

/*
 * Define a local structure to access the bit values of a byte.
 */
typedef struct tagBitArray
{
    int iBit1: 1;
    int iBit2: 1;
    int iBit3: 1;
    int iBit4: 1;
    int iBit5: 1;
    int iBit6: 1;
    int iBit7: 1;
    int iBit8: 1;
} BitArray, * pBitArray;

/*
 * Structure for "double" (i.e., 8-byte) integer.
 */
typedef struct tagDoubleInt
{
    long lLo;
    long lHi;
} DoubleInt, * pDoubleInt;

/*
 * Define constant for "double" mantissa bit masks corresponding to the two
 * elements of the preceding structure (DoubleInt).
 */
#define D_MANT_MASK_HI  (0x00080000)
#define D_MANT_MASK_LO  (0x80000000)

/******************************************************************************
 * ShowHelp()
 *
 * This subroutine prints the help/syntax for the program.
 *
 * INPUT:   sName   Name of executable
 *
 * OUTPUT:  None
 *
 * RETURNS: None
 *
 * NOTE(S): None
 *
 */
void ShowHelp( char * sName )
{
    printf( "\nSyntax: %s -[i|l|f|d] x\n\n", sName );
}

/******************************************************************************
 * ShowBit()
 *
 * This subroutine prints either a "1" or a "0" to standard output depending
 * on the logical state of the parameter.
 *
 * INPUT:   iBit    "Bit" to check
 *
 * OUTPUT:  None
 *
 * RETURNS: None
 *
 * NOTE(S): None
 *
 */
void ShowBit( int iBit )
{
    if ( iBit )
    {
        printf( "1" );
    }
    else
    {
        printf( "0" );
    }
}

/******************************************************************************
 * ShowBin()
 *
 * This subroutine displays, in binary format, the internal representation of
 * a memory location, for a certain number of bytes.
 *
 * A character pointer is used to reference the memory location in order to
 * preserve byte ordering.
 *
 * INPUT:   pLoc        Pointer to memory location
 *          iNumBytes   Number of bytes to display
 *
 * OUTPUT:  None
 *
 * RETURNS: None
 *
 * NOTE(S): It is the caller's responsibility to ensure the display starts
 *          at a reasonable position to allow for readability.
 *
 */
void ShowBin( char * pLoc, int iNumBytes )
{
    /*
     * Local Variables
     */
    BitArray sBitA; /* Bit array to be displayed */
    int      i;     /* Index                     */

    /*
     * Initialize local variable
     */
    sBitA.iBit1 = 0;
    sBitA.iBit2 = 0;
    sBitA.iBit3 = 0;
    sBitA.iBit4 = 0;
    sBitA.iBit5 = 0;
    sBitA.iBit6 = 0;
    sBitA.iBit7 = 0;
    sBitA.iBit8 = 0;

    /*
     * Go through each byte in sequence
     */
    for ( i = 0; i < iNumBytes; i++ )
    {
        /*
         * Copy the byte we want into the bit array variable
         * so it is easier to read and display.
         */
        memcpy( &sBitA, &(pLoc[i]), 1 );

        /*
         * Print out the bit information
         */
        ShowBit( sBitA.iBit1 );
        ShowBit( sBitA.iBit2 );
        ShowBit( sBitA.iBit3 );
        ShowBit( sBitA.iBit4 );
        ShowBit( sBitA.iBit5 );
        ShowBit( sBitA.iBit6 );
        ShowBit( sBitA.iBit7 );
        ShowBit( sBitA.iBit8 );

        /*
         * Separate each byte with a space for easier reading.
         */
        printf( " " );
    }
}

/******************************************************************************
 * ShowInject()
 *
 * This subroutine displays the RTByteBlock injection sequence for the value
 * provided.
 *
 * INPUT:   pLoc        Pointer to memory location
 *          iNumBytes   Number of bytes to display
 *
 * OUTPUT:  None
 *
 * RETURNS: None
 *
 * NOTE(S): 1. This subroutine does its own formatting.
 *          2. The injection sequence always starts at position 1.
 *
 */
void ShowInject( char * pLoc, int iNumBytes )
{
    /*
     * Local Variables
     */
    int i;  /* Index */

    /*
     * Print header
     */
    printf( "\n\tInjection sequence is:\n" );

    /*
     * Go through each byte
     */
    for ( i = 0; i < iNumBytes; i++ )
    {
        /*
         * This prints the injection sequence, with the injected value as a
         * decimal number, and its hexadecimal representation as a comment
         * following the injection.
         */
        printf( "\t\tat: %d put: %d. \"0x%02X\".\n", i, pLoc[i]&0xFF, pLoc[i]&0xFF );
    }
}

/******************************************************************************
 * ShowFloat()
 *
 * This subroutine displays a 4-byte floating point value as an equation from
 * its internal IEEE representation.
 *
 * INPUT:   fVal    Floating point value
 *
 * OUTPUT:  None
 *
 * RETURNS: None
 *
 * NOTE(S): This subroutine does its own formatting.
 *
 */
void ShowFloat( float fVal )
{
    /*
     * Local Variables
     */
    long  iSign;     /* Sign, either 1 or -1                      */
    long  iVal;      /* Integer occupying the same memory as fVal */
    long  iMantissa; /* Mantissa part of fVal                     */
    long  iExponent; /* Exponent part of fVal                     */
    long  iCnt;      /* Counter for # of bits in mantissa         */
    long  iMask      /* Mask for curren mantissa bit lookup       */
          = 0x00400000; /* Start at 1st bit (MSb)                 */
    long  iDiv       /* Current divider, corresponds to bit       */
          = 2;          /* First bit is 1/2, so iDiv = 2          */
    float fMantissa  /* Floating point value of the mantissa      */
          = 1.0;        /* Initialized to implied value 1.0       */

    /*
     * Start by assigning an integer to the same memory location
     * as the fVal.  This simplifies notation afterwards as we
     * will no longer need to do this horrible type cast...
     */
    iVal = *(long*)((char*)&fVal);

    /*
     * We get the sign from the first bit of the number.
     * If the bit is set, then the sign is negative, else we are dealing with
     * a positive number.
     */
    iSign = (iVal & 0x80000000) ? -1 : 1;

    /*
     * The exponent is taken from bits 2 through 9, inclusive.
     * Once the un-needed bits are masked out, we right shift the value by the
     * number of bits in the mantissa (23) so that is it properly aligned to
     * give us the correct integer value.  Finally, we substract the exponent
     * bias (127) from the resulting value to get the real exponent.
     */
    iExponent = ((iVal & 0x7F800000) >> 23) - 127;

    /*
     * The remaining bits (10 through 32, inclusive) form the mantissa.
     * However, this is not translatable directly into a number as it is a
     * binary decimal value.  That is, the first bit is 1/2, the second is
     * 1/4, etc. We will therefore need to translate this...  We could either
     * flip all the bits and invert the result, but it seems easier and more
     * expedient (but not necessarily faster) to simply check each value one
     * by one.
     */
    iMantissa = (iVal & 0x007FFFFF);

    /*
     * So for each bit in the mantissa, starting with the most significant:
     */
    for ( iCnt = 0; iCnt < 23; iCnt++ )
    {
        /*
         * If the bit is set...
         */
        if ( (iMantissa & iMask) == iMask )
        {
            /*
             * Add the current value to the floating point mantissa.
             */
            fMantissa += ((float)1.0 / (float)iDiv);
        }

        /*
         * Move on to the next mask...
         */
        iMask = iMask >> 1;

        /*
         * and to the next divider.
         */
        iDiv *= 2;
    }

    /*
     * Print the result.
     */
    printf( "\n\tWhich is: %ld * %g * 2^%ld\n", iSign, fMantissa, iExponent );
}

/******************************************************************************
 * ShowDouble()
 *
 * This subroutine displays an 8-byte floating point value as an equation from
 * its internal IEEE representation.
 *
 * INPUT:   dVal    Floating point value
 *
 * OUTPUT:  None
 *
 * RETURNS: None
 *
 * NOTE(S): This subroutine does its own formatting.
 *
 */
void ShowDouble( double dVal )
{
    /*
     * Local Variables
     */
    long    iSign;      /* Sign, either 1 or -1                      */
    long    iCnt;       /* Counter for # of bits in mantissa         */

	DoubleInt sVal;		/* Integer occupying the same memory as fVal */

    long	iMantissa;  /* Mantissa part of fVal                     */
    long	iExponent;  /* Exponent part of fVal                     */
	long	iMask		/* Mask for current mantissa bit lookup      */
		= D_MANT_MASK_HI;
    double	dDiv        /* Current divider, corresponds to bit       */
        = 2.0;              /* First bit is 1/2, so dDiv = 2.0       */

    double  dMantissa   /* Floating point value of the mantissa      */
        = 1.0;              /* Initialized to implied value 1.0      */

    /*
     * Start by assigning a DoubleInt structure to the same memory location as
     * the dVal.  This simplifies notation afterwards as we will no longer need
	 * to do this horrible type cast...
     */
    sVal = *(pDoubleInt)((char*)&dVal);

    /*
     * We get the sign from the first bit of the number.
     * If the bit is set, then the sign is negative, else we are dealing with
     * a positive number.
     */
#ifdef _MSC_VER
    iSign = (sVal.lHi & 0x80000000) ? -1 : 1;
#else
    iSign = (sVal.lLo & 0x80000000) ? -1 : 1;
#endif

    /*
     * The exponent is taken from bits 2 through 12, inclusive.
     * Once the un-needed bits are masked out, we right shift the value by the
     * number of bits in the mantissa (52) minus the number of bits in the
	 * "second" 8-byte part (32) so that is it properly aligned to give us the
	 * correct integer value.  Finally, we substract the exponent bias (1023)
	 * from the resulting value to get the real exponent.
     */
#ifdef _MSC_VER
    iExponent = ((sVal.lHi & 0x7FF00000) >> 20) - 1023;
#else
    iExponent = ((sVal.lLo & 0x7FF00000) >> 20) - 1023;
#endif

    /*
     * The remaining bits (13 through 64, inclusive) form the mantissa.
     * However, this is not translatable directly into a number as it is a
     * binary decimal value.  That is, the first bit is 1/2, the second is
     * 1/4, etc. We will therefore need to translate this...  We could either
     * flip all the bits and invert the result, but it seems easier and more
     * expedient (but not necessarily faster) to simply check each value one
     * by one.
	 *
	 * Start with the first part of the mantissa
	 */
#ifdef _MSC_VER
	iMantissa = sVal.lHi;
#else
	iMantissa = sVal.lLo;
#endif
	iMask = D_MANT_MASK_HI;

    /*
     * So for each bit in the mantissa, starting with the most significant:
     */
    for ( iCnt = 0; iCnt < 52; iCnt++ )
    {
        /*
         * if we are at the point where we pass the 8-byte boundary, move
         * on to the low values...
         */
        if ( iCnt == 20 )
        {
#ifdef _MSC_VER
            iMantissa = sVal.lLo;
#else
            iMantissa = sVal.lHi;
#endif
            iMask = D_MANT_MASK_LO;
        }

        /*
         * If the bit is set...
         */
        if ( (iMantissa & iMask) == iMask )
        {
            /*
             * Add the current value to the floating point mantissa.
             */
            dMantissa += ((double)1.0 / dDiv);
        }

        /*
         * Move on to the next mask...
         */
        iMask = iMask >> 1;

        /*
         * and to the next divider.
         */
        dDiv *= (double)2.0;
    }

    /*
     * Print the result.
     */
    printf( "\n\tWhich is: %ld * %g * 2^%ld\n", iSign, dMantissa, iExponent );
}

/******************************************************************************
 * PrintHexDouble()
 *
 * This function returns a string containing the hex display of an 8-byte
 * (double) floating point number.
 *
 * INPUT:   dVal    Floating point value
 *
 * OUTPUT:  None
 *
 * RETURNS: Pointer to the string.
 *
 * NOTE(S): The string is allocated locally!!!
 *
 */
char * PrintHexDouble( double dVal )
{
    /*
     * Local Variables
     */
    DoubleInt sVal;          /* Integer occupying the same memory as fVal */

    static char sOutStr[24]; /* Output string                             */

    /*
     * Start by assigning a DoubleInt structure to the same memory location as
     * the dVal.  This simplifies notation afterwards as we will no longer need
     * to do this horrible type cast...
     */
    sVal = *(pDoubleInt)((char*)&dVal);

    /*
     * Create the string.
     */
#ifdef _MSC_VER
    sprintf( sOutStr, "%08X%08X", sVal.lHi, sVal.lLo );
#else
    sprintf( sOutStr, "%08X%08X", sVal.lLo, sVal.lHi );
#endif

    return( sOutStr );
}

/******************************************************************************
 * main()
 *
 * This function runs the whole program.  It accepts input from the command
 * line, print usage information, and displays results.
 *
 * INPUT:   argc    Number of arguments
 *          argv    Array of arguments as strings
 *
 * OUTPUT:  None
 *
 * RETURNS: 0 on succeful completion
 *          1 on error
 *
 * NOTE(S): This program should be upgraded to provide fr prompting of
 *          information.
 *
 */
int main( int argc, char ** argv )
{
    /*
     * Local Variables
     */
    int	   iResNum  /* Integer input number              */
        = 0;
    long   lResNum  /* Long integer input number         */
        = 0;
    float  fResNum  /* float (4-byte real) input number  */
        = 0.0;
    double dResNum  /* double (8-byte real) input number */
        = 0.0;

    /*
     * Make sure all needed arguments are needed.
     */
    if (( argc < 3 ) ||
        ( argv[1][0] != '-' ))
    {
        ShowHelp( argv[0] );
        exit(1);
    }

    /*
     * Process the number, depending on the specified type.
     */
    switch( argv[1][1] )
    {
        /*
         * Integer:
         */
        case 'i':
            iResNum = atoi( argv[2] );
            printf( "\nInteger number %d [0x%0*X] is ", iResNum, (2*sizeof(int)), iResNum );
            ShowBin( (char *) &iResNum, sizeof(int) );
            ShowInject( (char *) &iResNum, sizeof(int) );
            break;

        /*
         * Long integer:
         */
        case 'l':
            lResNum = atol( argv[2] );
            printf( "\nLong integer number %ld [0x%0*lX] is ", lResNum, (2*sizeof(long)), lResNum );
            ShowBin( (char *) &lResNum, sizeof(long) );
            ShowInject( (char *) &lResNum, sizeof(long) );
            break;

        /*
         * 4-byte real (float):
         */
        case 'f':
            fResNum = (float) atof( argv[2] );
            printf( "\nFloat number %g [0x%0*X] is ", fResNum, (2*sizeof(float)), *(int*)((char*)&fResNum) );
            ShowBin( (char *) &fResNum, sizeof(float) );
            ShowFloat( fResNum );
            ShowInject( (char *) &fResNum, sizeof(float) );
            break;

        /*
         * 8-byte real (double):
         */
        case 'd':
            dResNum = atof( argv[2] );
            printf( "\nDouble number %g [0x%*s] is\n    ", dResNum, (2*sizeof(double)), PrintHexDouble(dResNum) );
            ShowBin( (char *) &dResNum, sizeof(double) );
            ShowDouble( dResNum );
            ShowInject( (char *) &dResNum, sizeof(double) );
            break;

        /*
         * Then there's always those who'll try anything...
         */
        default:
            ShowHelp( argv[0] );
            exit(1);
    }

    printf( "\n" );

    return(0);
	
}
